public extern const string VERSION;

public class Application : Gtk.Application
{
    private Gtk.ApplicationWindow window;
    private int window_width = 800;
    private int window_height = 600;
    private bool window_is_maximized = false;
    private MockupView view;
    private bool do_autosave = false;

    private Mockup _mockup;
    private Mockup mockup
    {
        get { return _mockup; }
        set
        {
            if (_mockup != null)
                SignalHandler.disconnect_by_func (_mockup, null, this);
            _mockup = value;
            view.mockup = value;
        }
    }

    public Application ()
    {
        Object (application_id: "com.ubuntu.mockup", flags: ApplicationFlags.HANDLES_OPEN);
    }

    public override void startup ()
    {
        base.startup ();

        window = new Gtk.ApplicationWindow (this);
        window.configure_event.connect ((event) =>
        {
            if (!window_is_maximized)
            {
                window_width = event.width;
                window_height = event.height;
            }
            return false;
        });
        window.window_state_event.connect ((event) =>
        {
            window_is_maximized = (event.new_window_state & Gdk.WindowState.MAXIMIZED) != 0;
            return false;
        });

        var filename = Path.build_filename (Environment.get_user_cache_dir (), "mockup", "state");
        var state = new KeyFile ();
        try
        {
            state.load_from_file (filename, KeyFileFlags.NONE);
            if (state.has_group ("window"))
            {
                window_width = state.get_integer ("window", "width");
                window_height = state.get_integer ("window", "height");
                window_is_maximized = state.get_boolean ("window", "is-maximized");
            }
        }
        catch (Error e)
        {
        }

        window.set_default_size (window_width, window_height);
        if (window_is_maximized)
            window.maximize ();

        var vbox = new Gtk.Box (Gtk.Orientation.VERTICAL, 0);
        vbox.visible = true;
        window.add (vbox);

        var toolbar = new Gtk.Toolbar ();
        toolbar.visible = true;
        toolbar.get_style_context ().add_class (Gtk.STYLE_CLASS_PRIMARY_TOOLBAR);
        vbox.pack_start (toolbar, false, true, 0);

        var button = new Gtk.ToolButton.from_stock (Gtk.Stock.NEW);
        button.is_important = true;
        button.clicked.connect (() => { mockup = new Mockup (); do_autosave = true; });
        button.visible = true;
        toolbar.add (button);

        button = new Gtk.ToolButton.from_stock (Gtk.Stock.OPEN);
        button.is_important = true;
        button.clicked.connect (open_cb);
        button.visible = true;
        toolbar.add (button);

        var save_button = new Gtk.MenuToolButton.from_stock (Gtk.Stock.SAVE);
        save_button.is_important = true;
        save_button.clicked.connect (() => { save (_("Untitled.mockup")); });
        save_button.visible = true;
        toolbar.add (save_button);

        var menu = new Gtk.Menu ();
        save_button.menu = menu;

        var item = new Gtk.MenuItem.with_label (_("Export as PNG"));
        item.visible = true;
        item.activate.connect (() => { save (_("Untitled.png")); });
        menu.append (item);

        item = new Gtk.MenuItem.with_label (_("Export as SVG"));
        item.visible = true;
        item.activate.connect (() => { save (_("Untitled.svg")); });
        menu.append (item);

        view = new MockupView ();
        view.visible = true;
        vbox.pack_start (view, true, true, 0);
    }

    public override void activate ()
    {
        var filename = Path.build_filename (Environment.get_user_cache_dir (), "mockup", "autosave.mockup");
        try
        {
            string contents;
            FileUtils.get_contents (filename, out contents);
            mockup = new Mockup.from_string (contents);
            do_autosave = true;
        }
        catch (FileError e)
        {
            if (!(e is FileError.NOENT))
                warning ("Failed to autoload: %s", e.message);
            mockup = new Mockup ();
            do_autosave = true;
        }

        window.visible = true;
    }

    public override void open (File[] files, string hint)
    {
        var filename = files[0].get_path ();
        try
        {
            string contents;
            FileUtils.get_contents (filename, out contents);
            mockup = new Mockup.from_string (contents);
            do_autosave = false;
        }
        catch (FileError e)
        {
            stderr.printf ("%s\n", e.message);
            quit ();
            return;
        }

        window.visible = true;
    }

    public override void shutdown ()
    {
        if (do_autosave)
        {
            var dir = Path.build_filename (Environment.get_user_cache_dir (), "mockup");
            DirUtils.create_with_parents (dir, 0775);
            try
            {
                FileUtils.set_contents (Path.build_filename (dir, "autosave.mockup"), mockup.to_string ());
            }
            catch (FileError e)
            {
                warning ("Failed to autosave: %s", e.message);
            }
        }

        var filename = Path.build_filename (Environment.get_user_cache_dir (), "mockup", "state");
        var state = new KeyFile ();
        state.set_integer ("window", "width", window_width);
        state.set_integer ("window", "height", window_height);
        state.set_boolean ("window", "is-maximized", window_is_maximized);
        try
        {
            FileUtils.set_contents (filename, state.to_data ());
        }
        catch (Error e)
        {
        }

        base.shutdown ();
    }

    private void open_cb ()
    {
        // FIXME: Warn about saving current mockup

        var dialog = new Gtk.FileChooserDialog (_("Open Mockup"), window, Gtk.FileChooserAction.OPEN,
                                                Gtk.Stock.CANCEL, Gtk.ResponseType.CANCEL,
                                                Gtk.Stock.OPEN, Gtk.ResponseType.ACCEPT);
        var filter = new Gtk.FileFilter ();
        filter.set_filter_name (_("Mockup Files"));
        filter.add_pattern ("*.mockup");
        dialog.add_filter (filter);
        filter = new Gtk.FileFilter ();
        filter.set_filter_name (_("All Files"));
        filter.add_pattern ("*");
        dialog.add_filter (filter);
        var response = dialog.run ();

        if (response == Gtk.ResponseType.ACCEPT)
        {
            var filename = dialog.get_file ().get_path ();
            
            try
            {
                string contents;
                FileUtils.get_contents (filename, out contents);
                mockup = new Mockup.from_string (contents);
                do_autosave = false;
            }
            catch (FileError e)
            {
                warning ("Failed to open %s: %s", filename, e.message);
            }
        }

        dialog.destroy ();
    }

    private void save (string filename)
    {
        var dialog = new Gtk.FileChooserDialog (_("Save Mockup"), window, Gtk.FileChooserAction.SAVE,
                                                Gtk.Stock.CANCEL, Gtk.ResponseType.CANCEL,
                                                Gtk.Stock.SAVE, Gtk.ResponseType.ACCEPT);
        dialog.do_overwrite_confirmation = true;
        dialog.set_current_name (filename);
        dialog.set_current_folder (Environment.get_user_special_dir (UserDirectory.DOCUMENTS));

        var filter = new Gtk.FileFilter ();
        filter.set_filter_name (_("Mockup Files"));
        filter.add_pattern ("*.mockup");
        dialog.add_filter (filter);
        if (filename.has_suffix (".mockup"))
            dialog.set_filter (filter);

        filter = new Gtk.FileFilter ();
        filter.set_filter_name (_("SVG Files"));
        filter.add_pattern ("*.svg");
        dialog.add_filter (filter);
        if (filename.has_suffix (".svg"))
            dialog.set_filter (filter);

        filter = new Gtk.FileFilter ();
        filter.set_filter_name (_("PNG Files"));
        filter.add_pattern ("*.png");
        dialog.add_filter (filter);
        if (filename.has_suffix (".png"))
            dialog.set_filter (filter);

        filter = new Gtk.FileFilter ();
        filter.set_filter_name (_("All Files"));
        filter.add_pattern ("*");
        dialog.add_filter (filter);

        var response = dialog.run ();
        if (response == Gtk.ResponseType.ACCEPT)
        {
            var path = dialog.get_file ().get_path ();

            int x, y, width, height;
            mockup.get_limits (out x, out y, out width, out height);

            if (path.has_suffix (".svg"))
            {
                var surface = new Cairo.SvgSurface (path, width, height);
                var c = new Cairo.Context (surface);
                c.translate (-x, -y);
                mockup.draw (c);
            }
            else if (path.has_suffix (".png"))
            {
                var surface = new Cairo.ImageSurface (Cairo.Format.RGB24, width, height);
                var c = new Cairo.Context (surface);
                c.translate (-x, -y);
                mockup.draw (c);
                surface.write_to_png (path);
            }
            else
            {
                try
                {
                    FileUtils.set_contents (path, mockup.to_string ());
                }
                catch (FileError e)
                {
                    warning ("Failed to save: %s", e.message);
                }
            }
        }

        dialog.destroy ();
    }
}

private void cairo_rounded_rectangle (Cairo.Context c, double x, double y, double width, double height, double radius)
{
    var w = width - radius * 2;
    var h = height - radius * 2;
    var kappa = 0.5522847498 * radius;
    c.move_to (x + radius, y);
    c.rel_line_to (w, 0);
    c.rel_curve_to (kappa, 0, radius, radius - kappa, radius, radius);
    c.rel_line_to (0, h);
    c.rel_curve_to (0, kappa, kappa - radius, radius, -radius, radius);
    c.rel_line_to (-w, 0);
    c.rel_curve_to (-kappa, 0, -radius, kappa - radius, -radius, -radius);
    c.rel_line_to (0, -h);
    c.rel_curve_to (0, -kappa, radius - kappa, -radius, radius, -radius);
}

public static int main (string[] args)
{
    var show_version = false;

    var entries = new OptionEntry[2];
    entries[0].long_name = "version";
    entries[0].short_name = 'v';
    entries[0].arg_data = &show_version;
    entries[0].description = _("Show release version");
    var options = new OptionContext ("[FILE]");
    options.add_main_entries (entries, null);
    options.add_group (Gtk.get_option_group (false));
    try
    {
        options.parse (ref args);
    }
    catch (OptionError e)
    {
        stderr.printf ("%s\n", e.message);
        return Posix.EXIT_FAILURE;
    }

    if (show_version)
    {
        stderr.printf ("mockup %s\n", VERSION);
        return Posix.EXIT_SUCCESS;
    }

    Gtk.init (ref args);

    var app = new Application ();
    return app.run (args);
}
