public class Mockup : Object
{
    private RootItem root_item;
    public MockItem? selected_item = null;
    public MockItemControlPoint? selected_control_point = null;

    public signal void item_added (MockItem item);
    public signal void item_removed (MockItem item);
    public signal void selected_item_changed ();

    public Mockup ()
    {
        root_item = new RootItem ();
    }

    public Mockup.from_string (string text)
    {
        this ();

        var item_table = new HashTable<string, MockItem> (str_hash, str_equal);
        var line_num = 0;
        foreach (var line in text.split ("\n"))
        {
            line_num++;
            line = line.strip ();
            if (line == "" || line.has_prefix ("#"))
                continue;

            if (line.has_prefix ("+"))
            {
                var tokens = line.split (" ", 3);
                if (tokens.length < 2)
                {
                    warning ("Unknown object on line %d", line_num);
                    continue;
                }
                
                var id = tokens[0].substring (1);
                var type = tokens[1];

                if (item_table.contains (id))
                {
                    warning ("Duplicate id %s on line %d", id, line_num);
                    continue;
                }

                MockItem item;
                switch (type)
                {
                case "box":
                    item = new MockBox ();
                    break;
                case "button":
                    item = new MockButton ();
                    break;
                case "check":
                    item = new MockCheck ();
                    break;
                case "icon":
                    item = new MockIcon ();
                    break;
                case "label":
                    item = new MockLabel ();
                    break;
                case "note":
                    item = new MockNote ();
                    break;
                case "progress-bar":
                    item = new MockProgressBar ();
                    break;
                case "radio":
                    item = new MockRadio ();
                    break;
                case "scroll-bar":
                    item = new MockScrollBar ();
                    break;
                case "switch":
                    item = new MockSwitch ();
                    break;
                case "slider":
                    item = new MockSlider ();
                    break;
                case "window":
                    item = new MockWindow ();
                    break;
                default:
                    warning ("Unknown object type %s on line %d", type, line_num);
                    continue;
                }

                MockItem? parent = null;
                if (tokens.length == 3)
                {
                    parent = item_table.lookup (tokens[2]);
                    if (parent == null)
                    {
                        warning ("Unknown object id %s on line %d", tokens[2], line_num);
                        continue;
                    }
                }

                item_table.insert (id, item);
                add_item (item, parent);
            }
            else
            {
                var tokens = line.split (".", 2);
                if (tokens.length != 2)
                {
                    warning ("Invalid line %d", line_num);
                    continue;
                }
                var id = tokens[0];
                tokens = tokens[1].split (" ", 2);
                if (tokens.length != 2)
                {
                    warning ("Invalid line %d", line_num);
                    continue;
                }
                var name = tokens[0];
                var value = tokens[1];

                var item = item_table.lookup (id);
                if (item == null)
                {
                    warning ("Unknown object id %s on line %d", id, line_num);
                    continue;
                }

                var prop = item.get_property (name);
                if (prop == null)
                {
                    warning ("Unknown property id %s on line %d", name, line_num);
                    continue;
                }

                prop.set_from_string (value);
            }
        }
    }

    public void get_limits (out int x, out int y, out int width, out int height)
    {
        var left = int.MAX;
        var right = int.MIN;
        var top = int.MAX;
        var bottom = int.MIN;

        foreach_item ((item) =>
        {
            if (item.x < left)
                left = item.x;
            if (item.x + item.width > right)
                right = item.x + item.width;
            if (item.y < top)
                top = item.y;
            if (item.y + item.height > bottom)
                bottom = item.y + item.height;
        });

        if (right < left || bottom < top)
        {
            x = y = width = height = 0;
            return;
        }

        x = left;
        width = right - left;
        y = top;
        height = bottom - top;
    }

    public void add_item (MockItem item, MockItem? parent = null)
    {
        if (parent == null)
            parent = root_item;
        item.parent = parent;
        parent.children.append (item);
        item_added (item);
    }
    
    public void foreach_item (MockItemFunc func, bool parent_to_child = true)
    {
        root_item.foreach_item (func, parent_to_child);
    }
   
    public bool select (int x, int y, out MockItem? item, out MockItemControlPoint? control_point)
    {
        item = null;
        control_point = null;

        /* Get the smallest item under the pointer */
        MockItem? selected_item = null;
        MockItemControlPoint? selected_control_point = null;
        foreach_item ((i) =>
        {
            foreach (var prop in i.properties)
            {
                if (!(prop is MockItemControlPoint))
                    continue;

                var p = prop as MockItemControlPoint;

                var dx = x - p.x;
                var dy = y - p.y;
                if (dx*dx + dy*dy < p.radius*p.radius)
                {
                    selected_item = i;
                    selected_control_point = p;
                    return;
                }
            }

            if (selected_control_point != null)
                return;

            if (!i.select (x, y))
                return;

            if (selected_item != null && selected_item.area < i.area)
                return;

            selected_item = i;
        });
        item = selected_item;
        control_point = selected_control_point;

        return item != null;
    }

    public void select_item (MockItem? item, MockItemControlPoint? control_point = null)
    {
        if (selected_item == item && selected_control_point == control_point)
            return;
        selected_item = item;
        selected_control_point = control_point;

        selected_item_changed ();
    }

    public void move (int dx, int dy)
    {
        if (selected_item == null)
            return;
            
        if (selected_control_point != null)
        {
            selected_control_point.x += dx;
            selected_control_point.y += dy;
            selected_control_point.moved ();
            return;
        }

        /* Move this item and its children */
        selected_item.foreach_item ((i) =>
        {
            i.x += dx;
            i.y += dy;
        });

        /* Make this item a child of the smallest and highest item it is inside */
        MockItem? parent = root_item;
        foreach_item ((i) =>
        {
            if (i == selected_item)
                return;

            if (i.contains (selected_item))
            {
                if (parent != root_item && (parent.depth > i.depth || parent.area < i.area))
                    return;

                parent = i;
            }
        });
        selected_item.parent.children.remove (selected_item);
        selected_item.parent = parent;
        selected_item.parent.children.append (selected_item);
    }

    public void resize (int dx, int dy)
    {
        if (selected_item == null)
            return;

        /* Move this item and its children */
        selected_item.width += dx;
        if (selected_item.width < 10)
            selected_item.width = 10;
        selected_item.height += dy;
        if (selected_item.height < 10)
            selected_item.height = 10;
    }

    public void remove_item (MockItem item)
    {
        item.parent.children.remove (item);

        item_removed (item);
        if (selected_item == item)
            select_item (null);
    }

    public void draw (Cairo.Context c)
    {
        root_item.foreach_item ((i) => { i.draw (c); });
    }

    public string to_string ()
    {
        var string = "# Mockup file\n";

        var item_ids = new HashTable<MockItem, string> (direct_hash, direct_equal);
        var index = 0;
        foreach_item ((i) =>
        {
            item_ids.insert (i, "%d".printf (index));
            index++;
        });

        foreach_item ((item) =>
        {
            string type;
            if (item is RootItem)
                return;
            else if (item is MockBox)
                type = "box";
            else if (item is MockButton)
                type = "button";
            else if (item is MockCheck)
                type = "check";
            else if (item is MockIcon)
                type = "icon";
            else if (item is MockLabel)
                type = "label";
            else if (item is MockNote)
                type = "note";
            else if (item is MockProgressBar)
                type = "progress-bar";
            else if (item is MockRadio)
                type = "radio";
            else if (item is MockScrollBar)
                type = "scroll-bar";
            else if (item is MockSlider)
                type = "slider";
            else if (item is MockSwitch)
                type = "switch";
            else if (item is MockWindow)
                type = "window";
            else
            {
                warning ("Unknown item type");
                return;
            }

            var id = item_ids.lookup (item);
            if (item.parent != root_item)
                string += "+%s %s %s\n".printf (id, type, item_ids.lookup (item.parent));
            else
                string += "+%s %s\n".printf (id, type);
            foreach (var prop in item.properties)
                string += "%s.%s %s\n".printf (id, prop.id, prop.to_string ());

            index++;
        });

        return string;
    }
}

public delegate void MockItemFunc (MockItem item);

public abstract class MockItem
{
    public unowned MockItem? parent;
    public List<MockItem> children;
    public bool resizable = false;
    public List<MockItemProperty> properties;
    private MockItemIntegerProperty x_prop;
    private MockItemIntegerProperty y_prop;
    private MockItemIntegerProperty width_prop;
    private MockItemIntegerProperty height_prop;
    
    public int x
    {
        get { return x_prop.value; }
        set { x_prop.value = value; }
    }

    public int y
    {
        get { return y_prop.value; }
        set { y_prop.value = value; }
    }

    public int width
    {
        get { return width_prop.value; }
        set { width_prop.value = value; }
    }

    public int height
    {
        get { return height_prop.value; }
        set { height_prop.value = value; }
    }

    public int depth
    {
        get
        {
            var d = 0;
            for (var p = this; p.parent != null; p = p.parent)
                d++;
            return d;
        }
    }
    
    public int area
    {
        get { return width * height; }
    }

    public signal void redraw ();

    public MockItem ()
    {
        properties = new List<MockItemProperty> ();
        x_prop = new MockItemIntegerProperty ("x", "X", 0, int.MIN, int.MAX);
        properties.append (x_prop);
        y_prop = new MockItemIntegerProperty ("y", "Y", 0, int.MIN, int.MAX);
        properties.append (y_prop);
        width_prop = new MockItemIntegerProperty ("width", "Width", 10, int.MIN, int.MAX);
        properties.append (width_prop);
        height_prop = new MockItemIntegerProperty ("height", "Height", 10, int.MIN, int.MAX);
        properties.append (height_prop);
    }
    
    public MockItemProperty? get_property (string id)
    {
        foreach (var prop in properties)
            if (prop.id == id)
                return prop;
        return null;
    }

    public abstract MockItem copy ();

    public abstract void draw (Cairo.Context c);

    public virtual bool select (int x, int y)
    {
        var dx = x - this.x;
        var dy = y - this.y;
        return dx >= 0 && dy >= 0 && dx < width && dy < height;
    }

    protected void add_property (MockItemProperty property)
    {
        properties.append (property);
    }

    public bool contains (MockItem item)
    {
        return item.x >= x && (item.x + item.width) <= (x + width) && item.y >= y && (item.y + item.height) <= (y + height);
    }
   
    public void foreach_item (MockItemFunc func, bool parent_to_child = true)
    {
        if (parent_to_child)
            func (this);
        foreach (var i in children)
            i.foreach_item (func, parent_to_child);
        if (!parent_to_child)
            func (this);
    }
}

public class RootItem : MockItem
{
    public override MockItem copy ()
    {
        return new RootItem ();
    }

    public override void draw (Cairo.Context c)
    {
        return;
    }
}

public abstract class MockItemProperty
{
    public string id;
    public string description;

    public signal void value_changed ();
    
    public MockItemProperty (string id, string description)
    {
        this.id = id;
        this.description = description;
    }

    public abstract string to_string ();
    public abstract void set_from_string (string text);
}

public class MockItemStringProperty : MockItemProperty
{
    public string _value = "";
    public string value
    {
        get { return _value; }
        set
        {
            _value = value;
            value_changed ();
        }
    }

    public MockItemStringProperty (string id, string description, string value)
    {
        base (id, description);
        _value = value;
    }

    public override string to_string ()
    {
        return value;
    }

    public override void set_from_string (string text)
    {
        value = text;
    }
}

public class MockItemChoiceProperty : MockItemProperty
{
    public string _value = "";
    public string value
    {
        get { return _value; }
        set
        {
            _value = value;
            value_changed ();
        }
    }
    public List<Choice> choices;

    public MockItemChoiceProperty (string id, string description, string value)
    {
        base (id, description);
        choices = new List<Choice> ();
        _value = value;
    }

    public void add_choice (string id, string description)
    {
        choices.append (new Choice (id, description));
    }

    public override string to_string ()
    {
        return value;
    }

    public override void set_from_string (string text)
    {
        value = text;
    }
}

public class Choice
{
    public string id;
    public string label;

    public Choice (string id, string label)
    {
        this.id = id;
        this.label = label;
    }
}

public class MockItemBooleanProperty : MockItemProperty
{
    public bool _value = false;
    public bool value
    {
        get { return _value; }
        set
        {
            _value = value;
            value_changed ();
        }
    }
    
    public MockItemBooleanProperty (string id, string description, bool value)
    {
        base (id, description);
        _value = value;
    }

    public override string to_string ()
    {
        return value ? "true" : "false";
    }

    public override void set_from_string (string text)
    {
        value = text == "true";
    }
}

public class MockItemIntegerProperty : MockItemProperty
{
    public int _value = 0;
    public int value
    {
        get { return _value; }
        set
        {
            if (value < min)
                value = min;
            else if (value > max)
                value = max;
            _value = value;
            value_changed ();
        }
    }
    public int min;
    public int max;
    
    public MockItemIntegerProperty (string id, string description, int value, int min, int max)
    {
        base (id, description);
        _value = value;
        this.min = min;
        this.max = max;
    }

    public override string to_string ()
    {
        return value.to_string ();
    }

    public override void set_from_string (string text)
    {
        value = int.parse (text);
    }
}

public class MockItemControlPoint : MockItemProperty
{
    public int x;
    public int y;
    public double radius;

    public signal void moved ();

    public MockItemControlPoint (string id, int x, int y, double radius = 3.5)
    {
        base (id, "");
        this.x = x;
        this.y = y;
        this.radius = radius;
    }

    public override string to_string ()
    {
        return "%d %d".printf (x, y);
    }

    public override void set_from_string (string text)
    {
        var tokens = text.split (" ");
        if (tokens.length == 2)
        {
            x = int.parse (tokens[0]);
            y = int.parse (tokens[1]);
        }
    }
}
